/**
 * {弹窗组件}
 * @author yutent<yutent.io@gmail.com>
 * @date 2023/03/06 15:17:25
 */

import { css, html, Component, bind, styleMap } from '@bd/core'
import '../form/input.js'

let uniqueInstance = null // 缓存当前打开的alert/confirm/prompt类型的弹窗
let toastInstance = null // 缓存toast的实例

const LANG_TITLE = '提示'
const LANG_BTNS = ['取消', '确定']
// 要保证弹层唯一的类型
const UNIQUE_TYPES = ['alert', 'confirm', 'prompt']
const BUILDIN_TYPES = UNIQUE_TYPES.concat(['toast'])

class Layer extends Component {
  static animation = {}

  static props = {
    type: {
      type: String,
      default: null,
      observer(v) {
        this.#wrapped = !BUILDIN_TYPES.includes(v)
      }
    },
    left: { type: String, attribute: false },
    right: { type: String, attribute: false },
    top: { type: String, attribute: false },
    bottom: { type: String, attribute: false },
    background: { type: String, attribute: false },
    'mask-color': { type: String, attribute: false },
    mask: false,
    'mask-close': false,
    title: { type: String, default: '', attribute: false },
    content: { type: String, default: '', attribute: false },
    btns: []
  }

  static styles = [
    css`
      :host {
        display: none;
        justify-content: center;
        align-items: center;
        position: fixed;
        z-index: 65534;
        left: 0;
        top: 0;
        width: 100%;
        height: 0;
      }
      :host([type]) {
        display: flex;
      }
      :host([type='toast']),
      :host([type='common']) {
        .layer {
          position: absolute;
        }
      }

      .noselect {
        -webkit-touch-callout: none;
        user-select: none;
        img,
        a {
          -webkit-user-drag: none;
        }
      }
    `,

    css`
      .layer {
        flex: 0 auto;
        position: relative;
        z-index: 65535;
        border-radius: 3px;
        color: #666;
        font-size: 14px;
        background: rgba(255, 255, 255, 0.8);
        box-shadow: 0 5px 20px rgba(0, 0, 0, 0.3);
        transition: opacity 0.2s ease-in-out, left 0.2s ease-in-out,
          right 0.2s ease-in-out, top 0.2s ease-in-out, bottom 0.2s ease-in-out;

        &:active {
          z-index: 65536;
        }
      }
    `,

    /* 弹层样式 */
    css`
      .layer {
        &__title {
          display: flex;
          justify-content: space-between;
          align-items: center;
          width: 100%;
          height: 60px;
          padding: 15px;
          font-size: 16px;
          color: var(--color-dark-2);

          wc-icon {
            --size: 14px;

            &:hover {
              color: var(--color-red-1);
            }
          }
        }

        &__content {
          display: flex;
          position: relative;
          width: 100%;
          height: auto;
          min-height: 50px;
          word-break: break-all;
          word-wrap: break-word;

          ::slotted(&__input) {
            flex: 1;
          }

          ::slotted(&__toast) {
            flex-shrink: 0;
            flex: 1;
            display: flex;
            align-items: center;
            width: 300px;
            min-height: 40px;
            margin-bottom: 15px !important;
            padding: 0 10px !important;
            border-radius: 3px;
            font-weight: normal;
            text-indent: 8px;
            --size: 16px;
            color: var(--color-dark-1);
            box-shadow: 0 2px 12px rgba(0, 0, 0, 0.1);
          }

          ::slotted(&__toast + &__toast) {
            margin-top: 30px;
          }

          ::slotted(&__toast.style-info) {
            border: 1px solid #ebeef5;
            background: #edf2fc;
            color: var(--color-grey-3);
          }
          ::slotted(&__toast.style-success) {
            border: 1px solid #e1f3d8;
            background: #f0f9eb;
            color: var(--color-green-3);
          }
          ::slotted(&__toast.style-warning) {
            border: 1px solid #faebb4;
            background: #faecd8;
            color: var(--color-red-1);
          }
          ::slotted(&__toast.style-error) {
            border: 1px solid #f5c4c4;
            background: #fde2e2;
            color: var(--color-red-1);
          }
        }

        &__ctrl {
          display: flex;
          justify-content: flex-end;
          width: 100%;
          height: 60px;
          padding: 15px;
          line-height: 30px;
          font-size: 14px;
          color: #454545;
          text-align: right;

          button {
            min-width: 64px;
            height: 30px;
            padding: 0 10px;
            margin: 0 5px;
            border: 1px solid var(--color-plain-3);
            border-radius: 3px;
            white-space: nowrap;
            background: #fff;
            font-size: inherit;
            font-family: inherit;
            outline: none;
            color: inherit;

            &:hover {
              background: var(--color-plain-1);
            }

            &:active {
              border-color: var(--color-grey-1);
            }

            &:focus {
              box-shadow: 0 0 0 2px var(--color-plain-a);
            }

            &:last-child {
              color: #fff;
              background: var(--color-teal-2);
              border-color: transparent;

              &:hover {
                background: var(--color-teal-1);
              }
              &:active {
                background: var(--color-teal-3);
              }
              &:focus {
                box-shadow: 0 0 0 2px var(--color-teal-a);
              }
            }

            &::-moz-focus-inner {
              border: none;
            }
          }
        }
      }
    `,

    css`
      :host([mask]) {
        height: 100%;
        background: rgba(0, 0, 0, 0.2);
      }

      :host([type='alert']),
      :host([type='confirm']),
      :host([type='prompt']) {
        .layer {
          max-width: 600px;
          min-width: 300px;
          background: #fff;

          &__content {
            padding: 0 15px;
          }
        }
      }

      :host([type='toast']) {
        .layer {
          box-shadow: none;
          background: none;

          &__content {
            flex-direction: column;
            min-height: 40px;
          }
        }
      }

      :host([blurry]) {
        .layer {
          backdrop-filter: blur(5px);
        }
      }
    `
  ]

  #wrapped = false

  #resolve = null
  #reject = null

  constructor() {
    super()

    this.promise = new Promise((resolve, reject) => {
      this.#resolve = resolve
      this.#reject = reject
    })

    this.promise.host = this
  }

  #intercept(value) {
    if (this.intercept) {
      this.intercept(value, _ => {
        delete this.intercept
        this.#resolve(value)
        this.$animate(true)
        this.$refs.box.$animate(true).then(_ => this.close())
      })
    } else {
      this.#resolve(value)
      this.$animate(true)
      this.$refs.box.$animate(true).then(_ => this.close())
    }
  }

  #play() {
    switch (this.type) {
      case 'toast':
        let elem = this.lastElementChild
        elem._anim = elem.animate(
          [
            { marginTop: '-30px', opacity: 0 },
            { marginTop: '0', opacity: 1 }
          ],
          {
            duration: 200,
            fill: 'forwards'
          }
        )

        setTimeout(() => {
          elem._anim.reverse()
          elem._anim.onfinish = _ => {
            elem.remove()
            if (this.children.length === 0) {
              this.close()
              toastInstance = null
            }
          }
        }, 3000)
        break

      default:
        this.$animate()
        this.$refs.box.$animate()
        break
    }
  }

  mounted() {
    if (this.type === 'prompt') {
      this.$refs.input = this.firstElementChild
      bind(this.$refs.input, 'submit', ev => {
        this.#intercept(ev.target.value)
      })
    } else if (this.type === 'toast') {
      this.style.display = ''
    }

    // 有遮罩层时
    if (this.mask) {
      this.$on('click', ev => {
        let path = ev.composedPath()

        if (path[0] === ev.currentTarget) {
          if (UNIQUE_TYPES.includes(this.type)) {
            return
          }

          if (this['mask-close']) {
            if (this.#wrapped === false) {
              this.#reject()
            }
            this.$refs.box.$animate(true).then(_ => this.close())
          }
        }
      })

      if (this['mask-color']) {
        this.style.backgroundColor = this['mask-color']
      }
    }

    if (this.background) {
      this.$refs.box.style.backgroundColor = this.background
    }

    this.#play()
  }

  updated() {
    this.#play()
  }

  moveTo(obj = {}) {
    var css = ''
    for (var k in obj) {
      css += `${k}:${obj[k]};`
    }
    this.$refs.box.style.cssText += css
  }

  show() {
    if (this.#wrapped) {
      this.type = 'common'
    }
  }

  /**
   * 关闭实例
   */
  close() {
    //
    if (this.#wrapped) {
      this.type = null
      this.$emit('close')
    } else {
      // 不允许多开的类型, 需要清除
      if (UNIQUE_TYPES.includes(this.type)) {
        uniqueInstance = null
      }
      this.$emit('close')
      this.remove()
    }
  }

  // 按钮的点击事件
  handleBtnClick(ev) {
    if (ev.target.tagName === 'BUTTON') {
      let idx = +ev.target.dataset.idx || 0

      switch (this.type) {
        case 'alert':
          this.#intercept(null)
          break

        case 'confirm':
        case 'prompt':
          if (idx === 0) {
            this.#reject()
            this.$animate(true)
            this.$refs.box.$animate(true).then(_ => this.close())
          } else {
            let value = this.type === 'prompt' ? this.$refs.input.value : null
            this.#intercept(value)
          }
          break

        default:
          // 其他类型, 如有按钮, 直接交给拦截器处理
          this.#intercept(idx)
          break
      }
    }
  }

  render() {
    let { type, mask, left, right, top, bottom } = this
    let styles = ''

    if (type === 'common' && mask === false) {
      top = top || 0
    }

    styles = styleMap({ left, right, top, bottom })

    return html`
      <div
        ref="box"
        class="layer"
        #animation=${{ type: 'micro-bounce' }}
        style=${styles}
      >
        <div
          class="layer__title noselect"
          style=${styleMap({ display: !!this.title ? '' : 'none' })}
        >
          ${this.title}
        </div>
        <div class="layer__content">
          <slot></slot>
        </div>
        <div
          class="layer__ctrl noselect"
          style=${styleMap({ display: this.btns.length ? '' : 'none' })}
          @click=${this.handleBtnClick}
        >
          ${this.btns.map((s, i) => html`<button data-idx=${i}>${s}</button>`)}
        </div>
      </div>
    `
  }
}

function layer(opt = {}) {
  let { type = 'common', content = '' } = opt
  let layDom =
    type === 'toast'
      ? toastInstance || document.createElement('wc-layer')
      : document.createElement('wc-layer')
  let alreadyInTree = type === 'toast' && !!toastInstance

  layDom.type = opt.type

  if (type === 'toast') {
    toastInstance = layDom
    layDom.top = '20px'
  } else {
    if (opt.btns && opt.btns.length) {
      layDom.btns = opt.btns
    }

    if (opt.intercept && typeof opt.intercept === 'function') {
      layDom.intercept = opt.intercept
    }

    layDom.mask = opt.mask
    layDom.title = opt.title

    // 这3种类型, 只允许同时存在1个, 如果之前有弹出则关闭
    if (UNIQUE_TYPES.includes(type)) {
      if (uniqueInstance) {
        uniqueInstance.$animate(true).then(_ => {
          uniqueInstance.close()
          uniqueInstance = layDom
        })
      } else {
        uniqueInstance = layDom
      }
    }
  }

  if (alreadyInTree) {
    let tmp = document.createElement('template')
    tmp.innerHTML = content

    layDom.appendChild(tmp.content.cloneNode(true))
    layDom.updated()
  } else {
    layDom.innerHTML = content
    document.body.appendChild(layDom)
  }

  return layDom.promise
}

layer.alert = function (
  content,
  title = LANG_TITLE,
  btns = LANG_BTNS.slice(1)
) {
  if (typeof title === 'object') {
    btns = title
    title = LANG_TITLE
  }
  return this({
    type: 'alert',
    title,
    content,
    mask: true,
    btns
  })
}

layer.confirm = function (
  content,
  title = LANG_TITLE,
  btns = LANG_BTNS.concat()
) {
  if (typeof title === 'object') {
    btns = title
    title = LANG_TITLE
  }
  return this({
    type: 'confirm',
    title,
    content,
    mask: true,
    btns
  })
}

layer.prompt = function (title = LANG_TITLE, defaultValue = '', intercept) {
  if (typeof defaultValue === 'function') {
    intercept = defaultValue
    defaultValue = ''
  }

  if (!intercept) {
    intercept = function (val, done) {
      if (val) {
        done()
      }
    }
  }
  return this({
    type: 'prompt',
    title,
    content: `<wc-input autofocus class="layer__content__input" value="${defaultValue}"></wc-input>`,
    mask: true,
    intercept,
    btns: LANG_BTNS.concat()
  })
}

layer.toast = function (txt, type = 'info') {
  var ico = type
  switch (type) {
    case 'info':
    case 'warning':
      break
    case 'error':
      ico = 'deny'
      break
    case 'success':
      ico = 'get'
      break
    default:
      ico = 'info'
  }

  return this({
    content: `
      <div class="layer__content__toast style-${type}">
        <wc-icon name="${ico}"></wc-icon>
        <span class="toast-txt">${txt}</span>
      </div>`,
    type: 'toast'
  })
}

Layer.reg('layer')
window.layer = layer
