/**
 * {选项卡组件}
 * @author yutent<yutent.io@gmail.com>
 * @date 2023/03/06 15:17:25
 */

import {
  css,
  html,
  bind,
  Component,
  nextTick,
  styleMap,
  classMap
} from '@bd/core'
import '../icon/index.js'

class Tabs extends Component {
  static props = {
    tab: { type: String, attribute: false },
    type: 'default',
    'tab-position': 'top',
    labels: []
  }

  static styles = [
    css`
      :host {
        display: flex;
        font-size: 14px;
      }

      :host([tab-position='top']) {
        flex-direction: column;
      }
      :host([tab-position='bottom']) {
        flex-direction: column-reverse;
      }

      :host([tab-position='right']) {
        flex-direction: row-reverse;
      }
      .ell {
        overflow: hidden;
        white-space: nowrap;
        text-overflow: ellipsis;
      }

      .header {
        position: relative;
        display: flex;
        width: 100%;
        height: 38px;
        color: var(--color-dark-1);
        user-select: none;
      }

      .navs {
        display: flex;
      }

      .content {
        flex: 1;
        padding: 16px 0;
      }

      .label {
        position: relative;
        display: inline-flex;
        align-items: center;
        justify-content: space-between;
        max-width: 120px;
        margin: 0 16px;
        --size: 16px;
        cursor: pointer;

        [slot='label'] {
          display: inline-flex;
          align-items: center;
        }

        &:first-child {
          margin-left: 0;
        }
        &:last-child {
          margin-right: 0;
        }

        &:hover:not([disabled]),
        &.active {
          color: var(--color-teal-1);
        }

        &[disabled] {
          opacity: 0.6;
          cursor: not-allowed;
        }

        .close {
          display: none;
          flex-shrink: 0;
          width: 14px;
          height: 14px;
          margin-left: 4px;
          padding: 3px;
          border-radius: 50px;
          color: var(--color-dark-1);
          transition: background 0.2s ease-in, color 0.2s ease-in;
        }

        &[closable='true'] {
          .close:hover {
            background: var(--color-red-1);
            color: #fff;
          }

          &:hover .close {
            display: inline-flex;
          }
        }
      }
    `,

    // type default
    css`
      :host([type='default']) {
        .header {
          border-bottom: 2px solid var(--color-plain-3);
        }

        .active-bar {
          position: absolute;
          left: 0;
          bottom: -2px;
          height: 2px;
          background: var(--color-teal-1);
        }
        .active-bar {
          transition: width 0.2s ease-out, left 0.2s ease-out, top 0.2s ease-out;
        }

        &:host([tab-position='bottom']) {
          .header {
            border-top: 2px solid var(--color-plain-3);
            border-bottom: 0;
          }
          .active-bar {
            top: -2px;
            bottom: auto;
          }
        }

        &:host([tab-position='left']),
        &:host([tab-position='right']) {
          .header {
            flex-direction: column;
            width: auto;
            height: auto;
            max-width: 120px;
            border-bottom: 0;
          }
          .navs {
            flex-direction: column;
          }
          .content {
            padding: 0 16px;
          }
          .label {
            margin: 0;
            height: 38px;
            padding: 0 16px;
          }
          .active-bar {
            width: 2px;
            height: 38px;
            bottom: auto;
          }
        }
        &:host([tab-position='left']) {
          .header {
            border-right: 2px solid var(--color-plain-3);
          }
          .label {
            justify-content: flex-end;
          }
          .active-bar {
            left: unset;
            right: -2px;
          }
        }
        &:host([tab-position='right']) {
          .header {
            border-left: 2px solid var(--color-plain-3);
          }

          .active-bar {
            left: -2px;
          }
        }
      }
    `,
    // type folder && card
    css`
      :host([type='folder']),
      :host([type='card']) {
        .content {
          padding: 16px;
        }

        .active-bar {
          display: none;
        }

        .label {
          margin: 0;
          padding: 0 16px;
          background: var(--color-plain-1);
        }
      }
    `,
    // type folder
    css`
      :host([type='folder']) {
        .header {
          overflow: hidden;
          padding: 0 6px;
        }

        .content {
          background: var(--tab-folder-color, #fff);
          box-shadow: 0 0 6px rgba(0, 0, 0, 0.1);
        }

        .label {
          padding-top: 2px;
          background: none;

          > span,
          > [slot='label'] {
            position: relative;
            z-index: 1;
          }

          &::before {
            position: absolute;
            top: 0;
            left: 0;
            z-index: -1;
            width: 100%;
            height: 36px;
            background: var(--color-plain-3);
            border-radius: 6px 6px 0 0;
            transform: perspective(9px) rotateX(3deg);
            box-shadow: 0 1px 6px rgba(0, 0, 0, 0.1);
            content: '';
          }

          &.active::before {
            z-index: 0;
            background: #fff;
          }
        }

        &:host([tab-position='bottom']) {
          .label {
            padding-top: 0;

            &::before {
              border-radius: 0 0 6px 6px;
              transform: perspective(9px) rotateX(-3deg);
              box-shadow: 0 -1px 6px rgba(0, 0, 0, 0.1);
            }
          }
        }

        &:host([tab-position='left']),
        &:host([tab-position='right']) {
          .header {
            flex-direction: column;
            width: auto;
            height: auto;
            padding: 4px 0;
          }
          .navs {
            flex-direction: column;
          }
          .label {
            height: 38px;
            padding-top: 0;
          }
        }

        &:host([tab-position='left']) {
          .label {
            justify-content: flex-end;

            &::before {
              border-radius: 6px 0 0 6px;
              transform: perspective(120px) rotateY(-20deg) translateX(-6px);
              box-shadow: -1px 0 6px rgba(0, 0, 0, 0.1);
            }
          }
        }
        &:host([tab-position='right']) {
          .label {
            &::before {
              border-radius: 0 6px 6px 0;
              transform: perspective(120px) rotateY(20deg) translateX(6px);
              box-shadow: -1px 0 6px rgba(0, 0, 0, 0.1);
            }
          }
        }
      }
    `,
    // type card
    css`
      :host([type='card']) {
        border: 1px solid var(--color-plain-3);
        background: #fff;

        .header {
          background: var(--color-plain-1);
          box-shadow: inset 0 -1px 0 0px var(--color-plain-3);
        }
        .label {
          border-right: 1px solid var(--color-plain-3);
          border-bottom: 1px solid var(--color-plain-3);

          &.active {
            border-bottom-color: transparent;
            background: #fff;
          }
        }

        &:host([tab-position='bottom']) {
          .header {
            box-shadow: inset 0 1px 0 0px var(--color-plain-3);
          }
          .label {
            border-top: 1px solid var(--color-plain-3);
            border-bottom: 0;
            &.active {
              border-top-color: transparent;
            }
          }
        }

        &:host([tab-position='left']),
        &:host([tab-position='right']) {
          .header {
            flex-direction: column;
            width: auto;
            height: auto;
          }
          .navs {
            flex-direction: column;
          }
          .label {
            height: 38px;
            padding-top: 0;

            &:last-child {
              border-bottom: 0;
            }

            &.active {
              background: #fff;
            }
          }
        }

        &:host([tab-position='left']) {
          .label {
            &.active {
              border-bottom-color: var(--color-plain-3);
              border-right-color: transparent;
            }
          }
        }
        &:host([tab-position='right']) {
          .label {
            border-left: 1px solid var(--color-plain-3);
            border-right: 0;

            &.active {
              border-bottom-color: var(--color-plain-3);
              border-left-color: transparent;
            }
          }
        }
      }
    `
  ]

  #cache = {}
  #tabWidth = 0 // 选项卡的宽度
  #tabLeft = 0 // 选项卡的左侧距离
  #tabTop = 0 // 选项卡的顶部距离

  // 处理子组件的slot,穿透到父组件来
  #parseSlot() {
    let left = 0
    let gaps = this.type === 'default' ? 32 : 0
    let top = 0
    this.labels.forEach((it, i) => {
      let nav = this.$refs.navs.children[i]
      if (it.el) {
        nav.replaceChild(it.el, nav.firstElementChild)
      }
      if (this.tab === it.name) {
        if (gaps) {
          this.#tabWidth = nav.clientWidth
          this.#tabLeft = left
          this.#tabTop = top
        }
        this.$requestUpdate()
        it.panel.$animate()
      }

      top += 38
      left += nav.clientWidth + gaps

      delete it.el
    })
  }

  selectTab(ev) {
    let elem = ev.target
    let key
    let left = 0
    let gaps = this.type === 'default' ? 32 : 0
    let top = 0
    if (elem === ev.currentTarget) {
      return
    }

    while (elem.tagName !== 'LABEL') {
      elem = elem.parentNode
    }

    key = this.labels[+elem.dataset.i].name

    if (key === this.tab) {
      return
    }

    if (this.labels[+elem.dataset.i].disabled) {
      return
    }

    // 非默认类型, 不显示active-bar
    if (gaps) {
      this.#tabWidth = elem.clientWidth

      for (let i = -1, it; (it = this.labels[++i]); ) {
        let nav = this.$refs.navs.children[i]
        if (key === it.name) {
          this.#tabLeft = left
          this.#tabTop = top
          break
        }
        top += 38
        left += nav.clientWidth + gaps
      }
    }

    this.#cache[this.tab].panel
      .$animate(true)
      .then(_ => this.#cache[key].panel.$animate())

    this.tab = key

    this.$emit('tab-change', {
      data: { name: key, label: this.#cache[key].label }
    })
  }

  removeTab(ev) {
    let tab = this.labels[+ev.target.dataset.i]
    this.$emit('tab-close', { data: { name: tab.name, label: tab.label } })
  }

  created() {
    bind(this.root, 'slotchange', ev => {
      let children = ev.target.assignedNodes()

      this.labels = children.map((it, i) => {
        let tmp = {
          label: it.label,
          name: (it.name || i).toString(),
          el: null,
          panel: null,
          disabled: it.disabled,
          closable: it.closable
        }
        this.#cache[tmp.name] = tmp
        return tmp
      })

      children.forEach((it, i) => {
        let slot = it.querySelector('[slot=label]')
        if (slot) {
          this.labels[i].el = slot.cloneNode(true)
          slot.remove()
        }
        this.labels[i].panel = it
      })

      this.tab = this.tab || this.labels[0].name

      nextTick(_ => this.#parseSlot())
    })
  }

  mounted() {
    //
  }

  render() {
    let tabStyle = {}
    if (this['tab-position'] === 'top' || this['tab-position'] === 'bottom') {
      tabStyle = {
        width: this.#tabWidth + 'px',
        left: this.#tabLeft + 'px'
      }
    } else {
      tabStyle = { top: this.#tabTop + 'px' }
    }

    return html`
      <header class="header">
        <section class="navs" ref="navs" @click=${this.selectTab}>
          ${this.labels.map(
            (it, i) =>
              html`<label
                class=${classMap({
                  label: true,
                  active: it.name === this.tab
                })}
                disabled=${it.disabled}
                closable=${it.closable}
                title=${it.label}
                data-i=${i}
              >
                <span class="ell">${it.label}</span>
                <wc-icon
                  class="close"
                  name="close"
                  data-i=${i}
                  @click.stop=${this.removeTab}
                ></wc-icon>
              </label>`
          )}
        </section>
        <span class="active-bar" style=${styleMap(tabStyle)}></span>
      </header>
      <div class="content">
        <slot></slot>
      </div>
    `
  }
}

class Tab extends Component {
  static animation = {}

  static props = {
    name: '',
    label: '',
    closable: false,
    disabled: false
  }
}

Tabs.reg('tabs')
Tab.reg('tab')
